/* (c) 1996-2001, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */

/* Tagfile Identification Database Interface
 *
 * This module supports using tagfiles as an identification database.  This
 * is a file named 'id' that sits in the user's home directory and has lines
 * like 'fullname=(Jan Wolter)'.  The directory path isn't stored in the
 * tagfile, because you need it before you can access the tagfile.  Instead
 * we generate the directory name from the login name.
 */


#include "backtalk.h"
#include "authident.h"

#include <ctype.h>
#include <errno.h>
#include <sys/stat.h>
#ifdef HAVE_SYS_FILE_H
#include <sys/file.h>
#endif

#include "set_uid.h"
#include "str.h"
#include "udb.h"
#include "lock.h"

char *showopt_ident_module= "ident_tagfile";

/* SAVE_TAG - Write the given variable/value pair into a tag file, replacing
 * the old value, if there was one, or appending the value.  The value is
 * not changed if it was protected, in which case a non-zero value is returned.
 * This always takes four arguments, but the last may be either an integer or
 * a string depending on the type.
 */

#if __STDC__
int save_tag(char *fname, int amadm, char *varname, int type, ...)
#else
int save_tag(fname, amadm, varname, type, va_alist)
    char *fname;
    int amadm;
    char *varname;
    int type;
    va_dcl
#endif
{
    va_list ap;
    char *bf, *val, *end, *var;
    FILE *fp;
    int i, protect, done= 0, rc= 0;

    bf= read_file(fname);
    VA_START(ap, type);

    /* Truncate and prepare to rewrite */
    if ((fp= fopen(fname,"w+")) == NULL)
	die("Cannot open file %s to write",fname);
    lock_exclusive(fileno(fp),fname);

    /* Process each line of saved file contents, writing back or replacing */
    for (var= bf; var != NULL && *var != '\0'; var= end+1)
    {
	/* Parse file line */
	if (protect= (*var == '*')) var++;
	val= strchr(var,'=');
	end= strchr(var,'\n');
	if (val != NULL && end != NULL && val > var && end > val)
	{
	    *(val++)= '\0';
	    *end= '\0';
	    if (!done && !strcmp(var,varname))
	    {
		done= 1;
		if (protect && !amadm)
		    rc= 1;
		else
		{
		    if (type == TK_INTEGER)
			fprintf(fp,"%s%s=%d\n", protect?"*":"",
			    var, va_arg(ap, int));
		    else
			fprintf(fp,"%s%s={%s}\n", protect?"*":"",
			    var, va_arg(ap, char *));
		    continue;
		}
	     }
	     fprintf(fp,"%s=%s\n", var-protect, val);
	}
    }

    /* Process the variable if not previously in file */
    if (!done)
    {
	if (type == TK_INTEGER)
	    fprintf(fp,"%s=%d\n", var, va_arg(ap, int));
	else
	    fprintf(fp,"%s={%s}\n", var, va_arg(ap, char *));
    }

    fclose(fp);
    unlock(fname);
    if (bf != NULL) free(bf);
    va_end(ap);
    return rc;
}


/* IDENT_GETPWNAM - Get the database entry for the named user.
 */

#ifdef HAVEIDENT_GETPWNAM
struct passwd *ident_getpwnam(char *nam)
{
    static struct passwd pw;
    static char fn[BFSZ];
    char fname[BFSZ+1];
    char bf[BFSZ+1], *dir, *var, *val, *end;
    FILE *fp;
    int ch;

    pw.pw_name= nam;

    /* Construct the directory name */
    if (nam == NULL || (dir= homedirpath(nam)) == NULL)
	return NULL;
    pw.pw_dir= dir;

    /* Open tag file */
    sprintf(fname,"%s/"IDENT_FILE,dir);
    if ((fp= fopen(fname,"r")) == NULL)
	    return NULL;
    lock_shared(fileno(fp),fname);
    
    /* Read through tag file */
    while (fgets(bf,BFSZ,fp) != NULL)
    {
	/* Parse apart the line */
	var= bf+ (bf[0] == '*');
	if ((val= strchr(var, '=')) == NULL)
	{
	    fclose(fp);
	    unlock(fname);
	    return NULL;
	}
	*(val++)= '\0';
	end= val + strlen(val) - 1;
	if (*end == '\n')
	    *(end--)= '\0';
	else
	    /* skip over rest of over-long line */
	    while ((ch= fgetc(fp)) != '\n' && ch != EOF)
		;

	if (!strcmp(var, "uid"))
	{
	    if (!isascii(*val) || !isdigit(*val))
	    {
		fclose(fp);
		unlock(fname);
		return NULL;
	    }
	    pw.pw_uid= atoi(val);
	}
	else if (!strcmp(var, "gid"))
	{
	    if (!isascii(*val) || !isdigit(*val))
	    {
		fclose(fp);
		unlock(fname);
		return NULL;
	    }
	    pw.pw_gid= atoi(val);
	}
	else if (!strcmp(var, "fullname"))
	{
	    /* We recognize both curly and round braces so both
	     * old and new file format works
	     */
	    if (*val != '(' && *val != '{')
	    {
		fclose(fp);
		unlock(fname);
		return NULL;
	    }
	    if (*end == ')' || *end == '}') *end= '\0';
	    unesctag(val+1,fn,0);
	    pw.pw_gecos= fn;
	}
    }

    fclose(fp);
    unlock(fname);
    return &pw;
}
#endif


/* IDENT_NEWFNAME - change the user's fullname to the given value.
 */

#ifdef HAVEIDENT_NEWFNAME
void ident_newfname(struct passwd *pw, char *newfname)
{
    char *fname;

    /* Construct name of ident file */
    fname= gethomedir(pw->pw_name);
    strcat(fname,"/");
    strcat(fname,IDENT_FILE);

    flushusercache();

    if (save_tag(fname, 0, "fullname", TK_STRING, newfname))
	die("Protected fullname field in %s",fname);
}
#endif


/* IDENT_NEWGID - change the user's groupid to the given value.
 */

#ifdef HAVEIDENT_NEWGID
void ident_newgid(struct passwd *pw, gid_t newgid)
{
    char *fname;

    /* Construct name of ident file */
    fname= gethomedir(pw->pw_name);
    strcat(fname,"/");
    strcat(fname,IDENT_FILE);

    flushusercache();

    save_tag(fname, 1, "gid", TK_INTEGER, newgid);
}
#endif


/* IDENT_ADD - Add an entry to the user database describing a user with
 * the given login, uid, gid, password, fullname, and directory.
 */

#ifdef HAVEIDENT_ADD
void ident_add(char *lid, uid_t uid, gid_t gid, char *fn, char *dir)
{
    char fname[BFSZ];
    FILE *fp;

    sprintf(fname,"%s/"IDENT_FILE,dir);
    if ((fp= fopen(fname,"w")) == NULL)
	die("could not create ident file %s",fname);
    lock_exclusive(fileno(fp),fname);
    fprintf(fp,"*uid=%d\n*gid=%d\nfullname={%s}\n", uid, gid, fn);
    fclose(fp);
    unlock(fname);
}
#endif


/* IDENT_DEL - Delete the named user from the ident database.  Actually, there
 * is nothing to do here, since we will always be deleting the directory
 * next, and that will take care of everything.
 */

#ifdef HAVEIDENT_DEL
void ident_del(char *name) { }
#endif
