/* Copyright 2001, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */


#include "backtalk.h"
#include "spell.h"

#if SPELLCHECK == 1

#include "sysdict.h"
#include "waittype.h"
#include "stack.h"
#include "str.h"
#include <ctype.h>


/* These are state information used by func_spellcheck.  If we wanted to
 * support nesting 'spellcheck' loops inside each other, these would have to
 * be on the execution stack, but I can't imagine wanting to do that,
 * so we don't bother.
 */

static FILE *spellin= NULL;	/* Pipe to spell checker */
static FILE *spellout= NULL;	/* Pipe from spell checker */
static pid_t spellcheck_pid;	/* Process ID of spell checker */
static int spell_start= 0;	/* Index of start of the current line */
static int spell_end= 0;	/* Index of end of the current line */
static int spell_intag= 0;	/* Are we currently in a tag? */
static int spell_inquote= 0;	/* Are we currently in quotes? */

/* These describe the current misspelled word */
static int guesses= 0;
char *guesslist= NULL;

static int ispell_offset;


/* OPEN_SPELLCHECK - start an ispell subprocess, setting up the spellin
 * and spellout global variables.
 */

void open_spellcheck()
{
    int pipe_in[2];
    int pipe_out[2];
    char bf[BFSZ+1];
    int ch;

    if (spellout != NULL) return;

    if (pipe(pipe_in) || pipe(pipe_out))
    	return;

    switch (spellcheck_pid= fork())
    {
    case 0: /* Child - do the exec */

	/* Close parent's ends of pipes */
	close(pipe_in[1]);
	close(pipe_out[0]);
	/* Attach pipe_in[0] to child's stdin */
	if (pipe_in[0] != 0)
	{
	    dup2(pipe_in[0],0);
	    close(pipe_in[0]);
	}
	/* Attach pipe_out[1] to child's stdout */
	if (pipe_out[1] != 1)
	{
	    dup2(pipe_out[1],1);
	    close(pipe_out[1]);
	}

#ifdef IISPELL_PATH
	if (sdict(VAR_SPELL_LANG)[0] == '\0')
	    execl(IISPELL_PATH, IISPELL_PATH, "-a", "-S", (char *)NULL);
	else
	    execl(IISPELL_PATH, IISPELL_PATH, "-a", "-S", "-d",
	    	sdict(VAR_SPELL_LANG),(char *)NULL);
#endif
	exit(127);

    case -1: /* Parent - Fork failed */
	close(pipe_in[0]);
	close(pipe_in[1]);
	close(pipe_out[0]);
	close(pipe_out[1]);
    	return;

    default: /* Parent - Fork succeeded */

	/* Close child's ends of the pipes */
    	close(pipe_in[0]);
    	close(pipe_out[1]);

	/* Start buffered I/O on useful ends of pipes */
	spellin= fdopen(pipe_in[1],"w");
	spellout= fdopen(pipe_out[0],"r");

	/* Put ispell in terse mode */
	fputs("!\n",spellin);
	fflush(spellin);

	/* Read & Discard Ispell's banner line */
	if ((ch= getc(spellout)) == EOF || ch != '@' ||
	    fgets(bf, BFSZ, spellout) == NULL)
	    die("Could not read ispell's banner line");

	/* Some versions of international ispell include the ^ when they report
	 * the offset of an error in a line, some don't.  We figure it out by
	 * trying it at run-time.
	 */
        fputs("^xfgzvwpqq\n",spellin);	/* some reasonably long non-word */
	fflush(spellin);
	if (fgets(bf,BFSZ,spellout) == NULL || getc(spellout) != '\n')
	    die("Could not communicate with ispell");
	ispell_offset= -atoi(firstin(bf,"0123456789"));
    }
}


/* CLOSE_SPELLCHECK - Close a spellcheck child process.  This must be the
 * most recently opened one.
 */

void close_spellcheck()
{
    pid_t wpid;

    if (spellout == NULL) return;
    fclose(spellout);
    if (spellin != NULL) fclose(spellin);
    if (guesslist != NULL) { free(guesslist); guesslist= NULL; }
    spellin= NULL;
    spellout= NULL;
    while ((wpid= wait((waittype *)NULL)) != -1 && wpid != spellcheck_pid)
    	;
}


/* SPELL_LINE - Given text, texttype, and an index, send the next line of
 * text to the spellchecker doing whatever filtering is necessary.  Returns
 * the index of the first character after the '\n' ending the line, or 0 if
 * there is no next line.
 */

int spell_line(char *text, int type, int begin)
{
    char ch;
    int i= begin;

    /* Quote each line with an initial ^ */
    putc('^',spellin);

    if (text[i] == '\0') return 0;

    for (;;)
    {
    	switch (ch= text[i]) {
	case '\n':
	    i++;
	case '\0':
	    putc('\n',spellin);
	    fflush(spellin);
	    return i;
	}
	i++;
	if (type == 1)
	{
	    if (spell_intag)
	    {
	    	if (spell_inquote)
		{
		   if (ch == spell_inquote)
		   	spell_inquote= 0;
		}
		else
		{
		    if (ch == '"' || ch == '\'')
		    	spell_inquote= ch;
		    else if (ch == '>')
		    	spell_intag= 0;
		}
		/* Convert insides of tags into blanks */
		ch= ' ';
	    }
	    else if (ch == '<')
	    	spell_intag= 1;
	    else if (ch == '&')
	    {
		/* Handle &nbsp; type tags (which are always on one line) */
		char *p= strchr(text+i,';');
	    	if (p != NULL && p < firstin(text+i," \n\r\t"))
		{
		    while (text+i <= p)
		    {
			i++;
		    	putc(' ',spellin);
		    }
		    ch= ' ';
		}
	    }
	}
	putc(ch,spellin);
    }
}


/* SPELLCHECK_STEP - Do the work for func_spellcheck.  We are given the text,
 * an index that points to the first character after the terminating newline
 * of the last line that we sent to 'ispell' , the mode, and the * texttype
 * (0=plain, 1=html).  Do the processing that has to happen before the
 * execution of the procedure.  If no mispellings are found, we return
 * non-zero.  If a match is found we
 *  - if mode is 1, push the section of the string before the match
 *  - push the misspelled word.
 *  - update the index.
 *  - return 0.
 */

int spellcheck_step(char *text, int *i, int mode, int type)
{
    char bf[BFSZ+1];
    char *p;
    int wordlen, offset;

    /* Check if we are starting a new spellcheck */
    if (*i == 0)
    {
    	open_spellcheck();
	spell_start= 0;
	spell_intag= 0;
	spell_inquote= 0;
	/* Send first line to ispell */
	if ((spell_end= spell_line(text,type,spell_start)) == 0) return 1;
    }

    while (fgets(bf,BFSZ,spellout))
    	switch(bf[0])
	{
	case '\n':	/* Line finished */
	    /* Feed next line of text to spellchecker */
	    spell_start= spell_end;
	    if ((spell_end= spell_line(text,type,spell_start)) == 0) return 1;
	    break;
	case '*':	/* Correctly spelled words */
	case '+':
	case '-':
	    break;

	case '&':	/* Incorrectly spelled words */
	case '?':
	case '#':

	    /* parse out misspelled word */
	    *(p= firstin(bf+2 ," \t\n"))= '\0';
	    wordlen= p - (bf + 2);
	    p++;

	    /* Parse out number of guesses */
	    if (bf[0] != '#')
	    {
	    	guesses= atoi(p);
		p= firstin(p," \t\n") + 1;
	    }
	    else
	    	guesses= 0;

	    /* Parse out starting offset of word */
	    offset= spell_start + atoi(p) + ispell_offset;

	    /* Parse out list of guesses */
	    if (guesses > 0)
	    {
	    	p= firstin(p,":") + 1;
		if (guesslist != NULL) free(guesslist);
		guesslist= strdup(p);
	    }
	    else
	    	guesslist= NULL;

	    /* Push correct block fragment */
	    if (mode == 1)
	    	push_string(strndup(text+(*i), offset-(*i)), FALSE);

	    /* Push misspelled word */
	    push_string(strndup(text+offset, wordlen), FALSE);

	    /* Update pointer to end of last correct string */
	    *i= offset + wordlen;

	    return 0;
	}
}

#endif /* SPELLCHECK */
